package org.qatools.validator;



import java.util.LinkedList;
import java.util.List;

import org.apache.log4j.Logger;


/**
 * <pre>
 * Scott Wilson presents QATools.org
 * Specializing in Functional & Integration Test Automation using Maven Java Junit TestNG
 * 
 * 
 * Validator extends Junit by capturing and reporting all failures at once.
 * This sppeds your QA effort by reducing the number of iterations.
 * 
 * Junit pukes at the first failed assertion and also does not enforce useful error messages
 * 			assertTrue(false);		<=== FAILS IMMEDIATELY
 * 			assertEquals(1, 2)		
 * 			assertNotNull(null);	
 * 
 * Validator example - the API forces you to enter useful error messages
 * 			Validator v = new Validator();
 * 			v.assertTrue("Foo msg", false);
 * 			v.assertEquals("Bar msg", 1, 2);
 * 			v.assertNotNull("Apple", null);
 * 			v.throwExceptionIfErrors();		<=== Captures ALL errors not just the first one!
 * 
 * </pre>
 * 
 * @author Scott Wilson - QATools.org
 */
public class AdvancedValidator extends AValidator {
    private static Logger logger = Logger.getLogger(AdvancedValidator.class);

	
	
    // ------------------------------------------------------
    // ----- class variables
    // ------------------------------------------------------

	//
	// Global static debug details
	private static List<String> sStaticDebugs = null;
	private static boolean sStaticDebuggingFlag = false;    // allows global debug messaging - must turn it on to use it
    private static boolean sStaticDebuggingAutoCleanOnMerge = true;    // automatically merge & clean static debugs
																		// client must resetStaticProperties if this property is false
    //TODO automatically associate static debugs to the TestCase rather than the Validator
    
    // ------------------------------------------------------
    // ----- constructors
    // ------------------------------------------------------

	public AdvancedValidator() {
        //do nothing constructor
    }
    


    
    // ----------------------------------------------------------------------------------------
    // ----- debugs - there are class & global static debugs
    // ----------------------------------------------------------------------------------------

    /**
     * Static debugging is OFF BY DEFAULT because some legacy frameworks using Validator do not exclusively use Validator. 
     * Test cases running only Junit might invoke utilities that do static debugging but these will not be cleaned up by test cases not invoking Validator assertions..
     * You must enable this feature and take responsibility for invoking Validator.resetStaticProperties() in your tearDown so the Junit ONLY test cases don't add out of scope static debug messages.
     */
    public static void disableStaticDebugging() {
        sStaticDebuggingFlag = false;        
    }
    /**
     * Static debugging is OFF BY DEFAULT because some legacy frameworks using Validator do not exclusively use Validator. 
     * Test cases running only Junit might invoke utilities that do static debugging but these will not be cleaned up by test cases not invoking Validator assertions..
     * You must enable this feature and take responsibility for invoking Validator.resetStaticProperties() in your tearDown so the Junit ONLY test cases don't add out of scope static debug messages.
     */
    public static void enableStaticDebugging() {
        sStaticDebuggingFlag = true;  
    }
    
    
    
    /**
     * Auto cleaning static debugging is enabled by default so debug messages are cleaned & do not pile up.
     * But this is not always desierable if you are using multiple validators. If the first validator merges & resets static debug message but has no errors,
     * then the second validator will lose the static debug messagets.
     * 
     * By dsiabling autoCleanOnMerge you can keep static debug messages after you assertions, but you must take responsibility
     * to lean them up manually be invoking Validator.resetStaticProperties()
     */
    public static void disableStaticDebuggingAutoCleanOnMerge() {
        sStaticDebuggingAutoCleanOnMerge = false;
    }
    /**
     * Auto cleaning static debugging is enabled by default so debug messages are cleaned & do not pile up.
     * But this is not always desierable if you are using multiple validators. If the first validator merges & resets static debug message but has no errors,
     * then the second validator will lose the static debug messagets.
     * 
     * By dsiabling autoCleanOnMerge you can keep static debug messages after you assertions, but you must take responsibility
     * to lean them up manually be invoking Validator.resetStaticProperties()
     */
    public static void enableStaticDebuggingAutoCleanOnMerge() {
        sStaticDebuggingAutoCleanOnMerge = true;        
    }
    
    
    /**
     * Returns the list of debug messages - CLASS & STATIC
     * 
     * @return
     */
    @Override
    public List<String> getDebugs() {
        if (debugs == null && sStaticDebugs == null) {
            return null;
            
        } else if (debugs != null && sStaticDebugs != null) {
            List<String> list = new LinkedList<String>();
            list.addAll (debugs);
            list.addAll (sStaticDebugs);
            return list;
            
        } else if (debugs != null) {
            return debugs;
        }
        return sStaticDebugs;
    }
//TODO not sure this is needed since there is also a mergeDebugs()
    
    
    

    
    /**
     * Adds an info to the debug list - STATIC GLOBAL LIST
     * 
     * Static debugging must be enabled otherwise it is not added to the list
     * 
     * @param error
     */
    public static void addStaticDebug(String info) {
        if (sStaticDebuggingFlag == false) {
            return;
        }
        //TODO - change the implementation to add it to the TestCase instead of the Validator
        //		but this might not work if the setUP is adding static debugs because the testcase is
        //		not known yet
        // maybe the getErrMsgIfExists() should move debugs to the testcase
        
        if (sStaticDebugs == null) {
            sStaticDebugs = new LinkedList<String>();
        }
        sStaticDebugs.add(info);
    }
    public static void addStaticDebug(String key, Object value) {
        addStaticDebug(key + " = " + value);
    }

    
    /**
     * Check if there are any debug messages
     * 
     * @return
     */
    @Override
    public boolean doDebugsExist() {
        if ( (debugs != null && 0 < debugs.size ()) || (sStaticDebugs != null && 0 < sStaticDebugs.size ()) ) {
            return true;
        }
        return false;
    }

    /**
     * Merge CLASS & STATIC debugs
     * 
     * Reset Static global debugs if sStaticDebuggingAutoCleanOnMerge == true
     */
    private void mergeDebugs() {
        if (sStaticDebugs != null && 0<sStaticDebugs.size ()) {
            if (debugs==null) {
                debugs = new LinkedList<String>();
            }
            debugs.addAll (sStaticDebugs);
            
            //
            // automatically clean up static debugs
            // you might want to disable this if you're using multiple validator methods
            // where the first one passes but the second one fails, you dont want the first
            // validator to clean up the static debugs
            if (sStaticDebuggingAutoCleanOnMerge == true) {
                sStaticDebugs = null;
            }
        }
    }
    
    
    
    // ------------------------------------------------------
    // ----- 
    // ------------------------------------------------------

    /**
     * Does not add detail to DetailReport
     * 
     * @param error
     */
    public void addError2(String error) {
    	super.addError(error);
    }
    
    /**
     * Adds detail to DetailReport
     */
    @Override
    public void addError(String error) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();
        DetailReport.recordFail(testCallerDetails, actualCallerDetails, error, error);
        addError2(error);
    }
    
    // ------------------------------------------------------
    // ----- 
    // ------------------------------------------------------

    
    /**
     * 
     * @param msg
     */
    public static void throwNewException(String msg) throws Exception {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        DetailReport.recordFail(testCallerDetails, actualCallerDetails, msg, msg);
        Validator v = new Validator();
        v.addError2(msg);
        v.throwExceptionIfErrors();
    }


    /**
     * 
     * @param msg
     */
    public static void throwNewException(Exception ex) throws Exception {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        String msg = null;
        if (ex == null) {
        	msg = "A NULL Exception was received";
        } else if (ex.getLocalizedMessage() != null) {
        	msg = ex.getLocalizedMessage();
        } else if (ex.getMessage() != null) {
        	msg = ex.getMessage();
        } else {
        	msg = "throwNewException";
        }
        DetailReport.recordFail(testCallerDetails, actualCallerDetails, ex.getLocalizedMessage (), ex.getLocalizedMessage ());
        Validator v = new Validator();
        v.addError2(msg);
        v.throwExceptionIfErrors();
    }
    

    /**
     * 
     * @param msg
     */
    public static void throwNewRuntimeException(String msg) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        DetailReport.recordFail(testCallerDetails, actualCallerDetails, msg, msg);
        Validator v = new Validator();
        v.addError2(msg);
        v.throwRuntimeExceptionIfErrors();
    }
    
    

    /**
     * 
     * @param msg
     */
    public static void throwNewRuntimeException(Exception ex) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        String msg = null;
        if (ex == null) {
        	msg = "A NULL Exception was received";
        } else if (ex.getLocalizedMessage() != null) {
        	msg = ex.getLocalizedMessage();
        } else if (ex.getMessage() != null) {
        	msg = ex.getMessage();
        } else {
        	msg = "throwNewRuntimeException";
        }
        DetailReport.recordFail(testCallerDetails, actualCallerDetails, ex.getLocalizedMessage (), ex.getLocalizedMessage ());
        Validator v = new Validator();
        v.addError2(msg);
        v.throwRuntimeExceptionIfErrors();
    }
    
    /**
     * Formats an error message for surefire & DetailReport.
     * Includes debug info.
     * 
     * @return
     */
    @Override
    protected String getErrorMsgIfExists() {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        mergeDebugs();
        //TODO this is where I could add debug info to the detail report

        String msg = null;
        if (doErrorsExist()) {          
                StringBuffer sb = new StringBuffer();

                //
                // check if the client only wants limited amount of info for the surefire report
                if (ErrorOption.SUREFIRE_SHORT.equals(sErrorOption)) {
                    sb.append("\n<br>" + errors.get(0));
                    sb.append("\n... more ... see target/QATools.org-detailReport.html");
                } else { //show all errors
                    formatErrors();
                    for (String str : errors) {
                        sb.append(str);
                    }
                }
    
                
                if (doDebugsExist()) {
                    DetailReport.addDebugInfo(testCallerDetails, debugs);	// all debugs added to DetailReport
                    if (DebugOption.EVERYWHERE.equals(sDebugOption)) {		// client may opt to limit debug info that shows in surefire
                        formatDebugs();
                        for (String str : debugs) {
                            sb.append(str);
                        }
                    }
                }
                sb.append("\n");
                msg = sb.toString();
        }  // if doErrorsExist
        
        //
        // 
        DetailReport.autoGenerateDetailReport(testCallerDetails);
        return msg;
    }
    
    
    // ------------------------------------------------------
    // ----- 
    // ------------------------------------------------------

    /**
     * Resets there error & debug lists
     */
    @Override
    public void reset() {
        errors = null;
        debugs = null;
        resetStaticProperties(); //not sure we want to reset these here
    }
    
    /**
     * Resets static debugs and 
     *         Validator.disableStaticDebugging();
     *         Validator.enableStaticDebuggingAutoCleanOnMerge();
     * 
     */
    public static void resetStaticProperties() {
        sStaticDebugs = null;
        AdvancedValidator.disableStaticDebugging();
        AdvancedValidator.enableStaticDebuggingAutoCleanOnMerge();
    }
    
    
    // ------------------------------------------------------
    // ----- Re-implements these Junit APIs
    // ------------------------------------------------------

    
	/**
	 * Asserts that a condition is true. If it isn't it adds the error to the error list..
	 */
	public void assertTrue(String message, boolean condition) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        if (!condition) {
        	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected True but was False");
			addError2(message + " expected True but was False");
		} else {
			DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected True and was True");
		}
	}
	

	
	/**
	 * Asserts that a condition is false. If it isn't it adds the error to the error list.
	 */
	public void assertFalse(String message, boolean condition) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

		if (condition) {
			addError2(message + " expected False but was True");
        	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected False but was True");
		} else {
			DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected False and was False");
		}
	}
	


	/**
	 * Adds the message to the error list
	 */
	public void fail(String message) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();
    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message);
		addError2(message);
	}



	/**
	 * Asserts that two objects are equal. If they are not
	 * add the message to the error list
	 */
	public void assertEquals(String message, Object expected, Object actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

		if (expected == null && actual == null) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected:'" + expected + "', actual:'" + actual + "'");
	    	return;
		}
		if (expected != null && expected.equals(actual)) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected:'" + expected + "', actual:'" + actual + "'");
			return;
		}
		addError2(formatErrorMsg(message, expected, actual));
    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, formatErrorMsg(message, expected, actual));
	}


	/**
	 * Asserts that two Strings are equal. 
	 */
	public void assertEquals(String message, String expected, String actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

		if (cleanString(expected) == null && cleanString(actual) == null) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected:'" + cleanStringForReport(expected) + "', actual:'" + cleanStringForReport(actual) + "'");
			return;
		}
		if (cleanString(expected) != null && cleanString(expected).equals(cleanString(actual))) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected:'" + cleanStringForReport(expected) + "', actual:'" + cleanStringForReport(actual) + "'");
			return;
		}
		addError2(formatErrorMsg(message, cleanStringForReport(expected), cleanStringForReport(actual)));
    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, formatErrorMsg(message, cleanStringForReport(expected), cleanStringForReport(actual)));
	}


	
	/**
	 * Asserts that two doubles are equal concerning a delta.  If the expected
	 * value is infinity then the delta value is ignored.
	 */
	public void assertEquals(String message, double expected, double actual, double delta) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

		// handle infinity specially since subtracting to infinite values gives NaN and the
		// the following test fails
		if (Double.isInfinite(expected)) {
			if (!(expected == actual)) {
				DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, formatErrorMsg(message, new Double(expected), new Double(actual)));
				addError2(formatErrorMsg(message, new Double(expected), new Double(actual)));
			}
		} else if (!(Math.abs(expected-actual) <= delta)) { // Because comparison with NaN always returns false
			DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, formatErrorMsg(message, new Double(expected), new Double(actual)));
			addError2(formatErrorMsg(message, new Double(expected), new Double(actual)));
		} else {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected:'" + expected + "', actual:'" + actual + "'");
		}
	}
	


	/**
	 * Asserts that two floats are equal concerning a delta. If they are not
	 * an AssertionFailedError is thrown with the given message.  If the expected
	 * value is infinity then the delta value is ignored.
	 */
	public void assertEquals(String message, float expected, float actual, float delta) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

 		// handle infinity specially since subtracting to infinite values gives NaN and the
		// the following test fails
		if (Float.isInfinite(expected)) {
			if (!(expected == actual)) {
				addError2(formatErrorMsg(message, new Float(expected), new Float(actual)));
				DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, formatErrorMsg(message, new Float(expected), new Float(actual)));
			}
		} else if (!(Math.abs(expected-actual) <= delta)) {
			addError2(formatErrorMsg(message, new Float(expected), new Float(actual)));			
			DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, formatErrorMsg(message, new Float(expected), new Float(actual)));
		} else {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected:'" + expected + "', actual:'" + actual + "'");
		}
	}

	
	/**
	 * Asserts that two longs are equal. If they are not
	 * an AssertionFailedError is thrown with the given message.
	 */
	public void assertEquals(String message, long expected, long actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

		if (expected != actual) {
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, formatErrorMsg(message, new Long(expected), new Long(actual)));
	    	addError2(formatErrorMsg(message, new Long(expected), new Long(actual)));
		} else
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected:'" + expected + "', actual:'" + actual + "'");
	}
	
	/**
	 * Asserts that two booleans are equal. If they are not
	 * an AssertionFailedError is thrown with the given message.
	 */
	public void assertEquals(String message, boolean expected, boolean actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

		if (expected != actual) {
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, formatErrorMsg(message, new Boolean(expected), new Boolean(actual)));
	    	addError2(formatErrorMsg(message, new Boolean(expected), new Boolean(actual)));
		} else {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected:'" + expected + "', actual:'" + actual + "'");
		}
  	}
	
	/**
	 * Asserts that two bytes are equal. If they are not
	 * an AssertionFailedError is thrown with the given message.
	 */
  	public void assertEquals(String message, byte expected, byte actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

		if (expected != actual) {
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, formatErrorMsg(message, new Byte(expected), new Byte(actual)));
	    	addError2(formatErrorMsg(message, new Byte(expected), new Byte(actual)));
		} else {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected:'" + expected + "', actual:'" + actual + "'");
		}
	}

  	
	/**
	 * Asserts that two chars are equal. If they are not
	 * an AssertionFailedError is thrown with the given message.
	 */
  	public void assertEquals(String message, char expected, char actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

		if (expected != actual) {
			addError2(formatErrorMsg(message, new Character(expected), new Character(actual)));
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, formatErrorMsg(message, new Character(expected), new Character(actual)));
		} else {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected:'" + expected + "', actual:'" + actual + "'");
		}
  	}

	/**
	 * Asserts that two shorts are equal. If they are not
	 * an AssertionFailedError is thrown with the given message.
	 */
	public void assertEquals(String message, short expected, short actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

		if (expected != actual) {
			addError2(formatErrorMsg(message, new Short(expected), new Short(actual)));
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, formatErrorMsg(message, new Short(expected), new Short(actual)));
		} else {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected:'" + expected + "', actual:'" + actual + "'");
		}
	}

	/**
	 * Asserts that two ints are equal. If they are not
	 * an AssertionFailedError is thrown with the given message.
	 */
  	public void assertEquals(String message, int expected, int actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

		if (expected != actual) {
			addError2(formatErrorMsg(message, new Integer(expected), new Integer(actual)));
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, formatErrorMsg(message, new Integer(expected), new Integer(actual)));
		} else {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected:'" + expected + "', actual:'" + actual + "'");
		}
  	}

	/**
	 * Asserts that an object isn't null. 
	 */
	public void assertNotNull(String message, Object object) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

		if (object == null) {
			addError2(message + " expected not null, but was null");
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected not null, but was null");
		} else {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected not null, and was not null");
		}
	}
	
	/**
	 * Asserts that an object is null. 
	 */
	public void assertNull(String message, Object object) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

		if (object != null) {
			addError2(message + " expected null, but was not null");
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected null, but was not null");
		} else {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected null, and was null");
		}
	}

	   /**
     * Asserts that an object is null. 
     */
    public void assertNull(String message, String object) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        if (cleanString(object) != null) {
            addError2(message + " expected null, but was not null - " + cleanStringForReport(object));
            DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected null, but was not null - " + cleanStringForReport(object));
        } else {
            DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected null, and was null");
        }
    }

	
    // ------------------------------------------------------
    // ----- 
    // ------------------------------------------------------

	/**
	 * asserts list is not null and not empty
	 * @param list
	 */
	public <T extends Object> void assertListNotNullNotEmpty(String message, List<T> list) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();
		
        if (list == null) {
			addError2(message + " expected notNullNotEmpty, but was null");
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected notNullNotEmpty, but was null");
        } else if (list.size() == 0) {
			addError2(message + " expected notNullNotEmpty, but was empty");
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected notNullNotEmpty, but was empty");
        } else {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected notNullNotEmpty, and was notNullNotEmpty");
        }        
	}
	
	/**
	 * asserts list is not null and not empty
	 * @param list
	 */
	public void assertStrNotNullNotEmpty(String message, String str) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();
		
        if (cleanString(str) == null) {
			addError2(message + " expected notNullNotEmpty, but was null - " + cleanStringForReport(str));
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected notNullNotEmpty, but was null - " + cleanStringForReport(str));
        } else if (cleanString(str).length() == 0) {
			addError2(message + " expected notNullNotEmpty, but was empty - " + cleanStringForReport(str));
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected notNullNotEmpty, but was empty - " + cleanStringForReport(str));
        } else {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected notNullNotEmpty, and was " + cleanStringForReport(str));
        }        
	}
	
    // ------------------------------------------------------
    // ----- Greater Than
    // ------------------------------------------------------

	/**
	 * actual should be greater than expected
	 * 
	 * @param message
	 * @param smallExpected
	 * @param bigActual
	 */
	public void assertGreaterThan(String message, long smallExpected, long bigActual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

       if (smallExpected < bigActual) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " " + smallExpected + " < " + bigActual);
       } else {
			addError2(message + " " + smallExpected + " NOT < " + bigActual);
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " " + smallExpected + " NOT < " + bigActual);
       }
	}
	/**
	 * actual should be greater than expected
	 * 
	 * @param message
	 * @param smallExpected
	 * @param bigActual
	 */
	public void assertGreaterThan(String message, int smallExpected, int bigActual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

       if (smallExpected < bigActual) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " " + smallExpected + " < " + bigActual);
       } else {
			addError2(message + " " + smallExpected + " NOT < " + bigActual);
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " " + smallExpected + " NOT < " + bigActual);
       }
	}

	
	/**
	 * actual should be greater than expected
	 * 
	 * @param message
	 * @param smallExpected
	 * @param bigActual
	 */
	public void assertGreaterThanEquals(String message, long smallExpected, long bigActual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

       if (smallExpected <= bigActual) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " " + smallExpected + " <= " + bigActual);
       } else {
			addError2(message + " " + smallExpected + " NOT <= " + bigActual);
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " " + smallExpected + " NOT <= " + bigActual);
       }
	}
	/**
	 * actual should be greater than expected
	 * 
	 * @param message
	 * @param smallExpected
	 * @param bigActual
	 */
	public void assertGreaterThanEquals(String message, int smallExpected, int bigActual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

       if (smallExpected <= bigActual) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " " + smallExpected + " <= " + bigActual);
       } else {
			addError2(message + " " + smallExpected + " NOT <= " + bigActual);
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " " + smallExpected + " NOT <= " + bigActual);
       }
	}

    // ------------------------------------------------------
    // ----- Less Than
    // ------------------------------------------------------

    /**
     * actual should be greater than expected
     * 
     * @param message
     * @param bigExpected
     * @param smallActual
     */
    public void assertLessThan(String message, long bigExpected, long smallActual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

       if (bigExpected > smallActual) {
            DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " " + bigExpected + " > " + smallActual);
       } else {
            addError2(message + " " + bigExpected + " NOT > " + smallActual);
            DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " " + bigExpected + " NOT > " + smallActual);
       }
    }
    /**
     * actual should be greater than expected
     * 
     * @param message
     * @param bigExpected
     * @param smallActual
     */
    public void assertLessThan(String message, int bigExpected, int smallActual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

       if (bigExpected > smallActual) {
            DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " " + bigExpected + " > " + smallActual);
       } else {
            addError2(message + " " + bigExpected + " NOT > " + smallActual);
            DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " " + bigExpected + " NOT > " + smallActual);
       }
    }

    
    /**
     * actual should be greater than expected
     * 
     * @param message
     * @param bigExpected
     * @param smallActual
     */
    public void assertLessThanEquals(String message, long bigExpected, long smallActual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

       if (bigExpected >= smallActual) {
            DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " " + bigExpected + " >= " + smallActual);
       } else {
            addError2(message + " " + bigExpected + " NOT >= " + smallActual);
            DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " " + bigExpected + " NOT >= " + smallActual);
       }
    }
    /**
     * actual should be greater than expected
     * 
     * @param message
     * @param bigExpected
     * @param smallActual
     */
    public void assertLessThanEquals(String message, int bigExpected, int smallActual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

       if (bigExpected >= smallActual) {
            DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " " + bigExpected + " >= " + smallActual);
       } else {
            addError2(message + " " + bigExpected + " NOT > " + smallActual);
            DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " " + bigExpected + " NOT >= " + smallActual);
       }
    }
    
    // ------------------------------------------------------
    // ----- 
    // ------------------------------------------------------

	/**
	 * The actual list should contain the expected object
	 * 
	 * @param expected
	 * @param actual
	 */
	public <T extends Object> void assertContains(String message, T expected, List<T> actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        if (actual==null) {
			addError2(message + " actual is NULL");
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " actual is NULL");
        	
        } else if (actual.size() == 0) {
			addError2(message + " actual size is zero");
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " actual size is zero");
        	
        } else if (actual.contains(expected)) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " actual '" + actual.toString() + "' contains expected '" + expected + "'");
        } else {
			addError2(message + " actual '" + actual.toString() + "' does NOT contain expected '" + expected + "'");
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " actual '" + actual.toString() + "' does NOT contain expected '" + expected + "'");
        }
	}
	
	/**
	 * The actual list should contain the expected object
	 * 
	 * @param expected
	 * @param actual
	 */
	public <T extends Object> void assertContains(String message, List<T> expected, T actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        if (expected==null) {
			addError2(message + " expected is NULL");
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected is NULL");
        	
        } else if (expected.size() == 0) {
			addError2(message + " expected size is zero");
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected size is zero");

        } else if (expected.contains(actual)) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected '" + expected.toString() + "' contains actual '" + actual + "'");
        } else {
			addError2(message + " expected '" + expected.toString() + "' does NOT contain actual '" + actual + "'");
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected '" + expected.toString() + "' does NOT contain actual '" + actual + "'");
        }
	}

	/**
	 * The actual string must be contained within the expected
	 * 
	 * @param message
	 * @param longExpected
	 * @param shortActual
	 */
	public void assertStrContainsStr(String message, String longExpected, String shortActual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        String tmpLongExpected = cleanString(longExpected);
        String tmpShortActual = cleanString(shortActual);
        
        if (tmpLongExpected==null && tmpShortActual==null) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected '" + cleanStringForReport(longExpected) + "' contains actual '" + cleanStringForReport(shortActual) + "'");
        	
        } else if (tmpLongExpected.contains(tmpShortActual)) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected '" + cleanStringForReport(longExpected) + "' contains actual '" + cleanStringForReport(shortActual) + "'");
        } else {
			addError2(message + " expected '" + longExpected + "' does NOT contain actual '" + shortActual + "'");
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected '" + cleanStringForReport(longExpected) + "' does NOT contain actual '" + cleanStringForReport(shortActual) + "'");
        }
	}

	/**
	 * The expected must be contained within the actual
	 * 
	 * @param message
	 * @param shortExpected
	 * @param longActual
	 */
	public void assertStrContainsStr2(String message, String shortExpected, String longActual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        String tmpShortExpected = cleanString(shortExpected);
        String tmpLongActual = cleanString(longActual);
         
        if (tmpLongActual==null && tmpShortExpected==null) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " actual '" + cleanStringForReport(longActual) + "' contains expected '" + cleanStringForReport(shortExpected) + "'");
        	
        } else if (tmpLongActual.contains(tmpShortExpected)) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " actual '" + cleanStringForReport(longActual) + "' contains expected '" + cleanStringForReport(shortExpected) + "'");
        } else {
			addError2(message + " actual '" + cleanStringForReport(longActual) + "' does NOT contain expected '" + cleanStringForReport(shortExpected) + "'");
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " actual '" + cleanStringForReport(longActual) + "' does NOT contain expected '" + cleanStringForReport(shortExpected) + "'");
        }
	}

	// ------------------------------------------------------
    // ----- 
    // ------------------------------------------------------

	/**
	 * 
	 * @param message
	 * @param list
	 * @param list2
	 */
	public void assertEquals(String message, List<Object> expected, List<Object> actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();
		
        if (expected==null && actual!=null) {
        	String msg = message + " expected is null, actual is not null(size:" + actual.size() + ")";
			addError2(msg);
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, msg, msg);
        	return;
        	
        } else if (expected!=null && actual==null) {
        	String msg = message + " expected is not null(size:" + expected.size() + ") actual is null)";
			addError2(msg);
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, msg, msg);
        	return;        	
        } 

        if (expected.size() != actual.size()) {
        	String msg = message + " expected.size=" + expected.size() + ", actual.size=" + actual.size();
			addError2(msg);
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, msg, msg);
	    	addDebug(message + ", expected=" + expected.toString());
	    	addDebug(message + ", actual=" + actual.toString());
        	return;        	        	
        }
        
        boolean flag = true;
        for (int i=0; i<expected.size(); i++) {
        	Object e = expected.get(i);
        	Object a = actual.get(i);
        	if (!e.equals(a)) {
        		flag = false;
        	}
        }
        if (flag) {
        	String msg = message + ", expected.list matches actual.list";
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, msg);
        } else {
        	String msg = message + ", expected.list does NOT match actual.list";
			addError2(msg);
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, msg, msg);
	    	addDebug(message + ", expected=" + expected.toString());
	    	addDebug(message + ", actual=" + actual.toString());        	
        }
        
	}
	
	// ------------------------------------------------------
    // ----- assertNotEquals
    // ------------------------------------------------------

	/**
	 * Asserts that two objects are NOT equal. If they are equal
	 * add the message to the error list
	 */
	public void assertNotEquals(String message, Object expected, Object actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        String msg = message + " - NOT expected:'" + expected + "', actual:'" + actual + "'";
		if (expected == null && actual == null) {
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, "NOT expected:'" + expected + "', actual:'" + actual + "'");
			addError2(msg);
	    	return;
		}
		if (expected != null && expected.equals(actual)) {
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, "NOT expected:'" + expected + "', actual:'" + actual + "'");
			addError2(msg);
			return;
		}
    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, msg);
	}


	/**
	 * Asserts that two Strings are NOT equal. 
	 */
	public void assertNotEquals(String message, String expected, String actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        String ex = cleanString(expected);
        String a = cleanString(actual);
        String msg = message + " - NOT expected:'" + ex + "', actual:'" + a + "'";

		if (ex == null && a == null) {
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, msg);
			addError2(msg);
			return;
		}
		if (ex != null && ex.equals(a)) {
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, msg);
			addError2(msg);
			return;
		}
    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, msg);
	}

	
	/**
	 * Asserts that two longs are equal. If they are not
	 * an AssertionFailedError is thrown with the given message.
	 */
	public void assertNotEquals(String message, long expected, long actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        String msg = message + " - NOT expected:'" + expected + "', actual:'" + actual + "'";
        
		if (expected != actual) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, msg);
		} else {
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, msg);
    		addError2(msg);
		}
	}
	
	/**
	 * Asserts that two booleans are equal. If they are not
	 * an AssertionFailedError is thrown with the given message.
	 */
	public void assertNotEquals(String message, boolean expected, boolean actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();
        String msg = message + " - NOT expected:'" + expected + "', actual:'" + actual + "'";

		if (expected != actual) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, msg);
		} else {
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, msg);
	    	addError2(msg);
		}
  	}
	

	/**
	 * Asserts that two ints are equal. If they are not
	 * an AssertionFailedError is thrown with the given message.
	 */
  	public void assertNotEquals(String message, int expected, int actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();
        String msg = message + " - NOT expected:'" + expected + "', actual:'" + actual + "'";

		if (expected != actual) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, msg);
		} else {
			addError2(msg);
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, msg);
		}
  	}
	
	// ------------------------------------------------------
    // ----- range
    // ------------------------------------------------------

  	/**
  	 * 
  	 * @param message
  	 * @param expectedLow
  	 * @param expectedHigh
  	 * @param actual
  	 */
	public void assertRange(String message, int expectedLow, int expectedHigh, int actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();
        String msg = message + " - expected range:'" + expectedLow + "-" + expectedHigh + "', actual:'" + actual + "'";

		if ((expectedLow <= actual) && (actual <= expectedHigh)) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, msg);
		} else {
			addError2(msg);
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, msg);
		}
	}

  	/**
  	 * 
  	 * @param message
  	 * @param expectedLow
  	 * @param expectedHigh
  	 * @param actual
  	 */
	public void assertRange(String message, long expectedLow, long expectedHigh, long actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();
        String msg = message + " - expected range:'" + expectedLow + "-" + expectedHigh + "', actual:'" + actual + "'";

		if ((expectedLow <= actual) && (actual <= expectedHigh)) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, msg);
		} else {
			addError2(msg);
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, msg);
		}
	}


	
	
}
